/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __operator_TransformManipulator_h__
#define __operator_TransformManipulator_h__

namespace fe
{

/**************************************************************************//**
	@brief Manipulator to adjust a SpatialTransform

	@ingroup operator
*//***************************************************************************/
class FE_DL_EXPORT TransformManipulator:
	public ManipulatorCommon,
	public Initialize<TransformManipulator>
{
	public:

					TransformManipulator(void):
						m_mode(e_deform)									{}
virtual				~TransformManipulator(void)								{}

		void		initialize(void);

					//* As HandlerI (window events)
virtual	void		handle(Record& a_rSignal);

static	Real		flameProfile(Real a_peakOffset,Real a_fraction);

	protected:
					//* As ManipulatorCommon
virtual	void		updateGrips(void);

	private:

		void		updateTransform(void);

		void		drawTubeWithAura(sp<DrawI>& a_rspDraw,
							const SpatialVector* a_point,
							const SpatialVector* a_normal,
							const Real* a_radius,
							U32 a_pointCount,CurveMode a_curveMode,
							BWORD a_multicolor,const Color *a_color,
							const SpatialVector& a_rTowardCamera,
							const Real a_auraScale);

		enum		Mode
					{
						e_deform,
						e_pivot
					};

		Mode				m_mode;
		sp<DrawMode>		m_spTubeAura;
		sp<DrawMode>		m_spCircleAura;

		SpatialTransform	m_unscaled;
		SpatialVector		m_scaling;
		SpatialVector		m_facing;
		SpatialVector		m_upwards;
		SpatialVector		m_rightways;
};

} /* namespace */

#endif /* __operator_TransformManipulator_h__ */
