/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __vdb_SurfaceVDB_h__
#define __vdb_SurfaceVDB_h__

namespace fe
{

/**************************************************************************//**
	@brief VDB Surface

	@ingroup vdb
*//***************************************************************************/
class FE_DL_EXPORT SurfaceVDB: public SurfaceSphere
{
	public:
	class Impact: public SurfaceSphere::Impact
	{
		public:
							Impact(void)
							{
#if FE_COUNTED_STORE_TRACKER
								setName("SurfaceVDB::Impact");
#endif
							}
	virtual					~Impact(void)									{}

	virtual	Vector2			uv(void)
							{	return m_uv; }
			void			setUV(Vector2 a_uv)
							{	m_uv=a_uv; }

		protected:
	virtual	void			resolve(void)
							{

								const SurfaceVDB* pSurfaceVDB=
										dynamic_cast<const SurfaceVDB*>(
										m_pSurface);
								pSurfaceVDB->resolveImpact(
										sp<ImpactI>(this));
								m_resolved=TRUE;
							}

			Vector2			m_uv;

	};
								SurfaceVDB(void);
virtual							~SurfaceVDB(void);

								//* As Protectable
virtual	Protectable*			clone(Protectable* pInstance=NULL);

								//* As SurfaceI
virtual	void					bind(Instance a_instance);

virtual Containment				containment(
										const SpatialVector& a_origin) const;

virtual	sp<SurfaceI::ImpactI>	nearestPoint(
										const SpatialVector& a_origin) const;

virtual	void					draw(const SpatialTransform&,
										sp<DrawI> a_spDrawI,
										const fe::Color* a_color) const;

	protected:

virtual	void					resolveImpact(sp<ImpactI> a_spImpactI) const {}

	private:

		void					bindGridBase(
										openvdb::GridBase::ConstPtr& a_gridPtr);

		openvdb::FloatGrid::ConstPtr	m_sdfGridPtr;
		openvdb::Vec3SGrid::ConstPtr	m_cptGridPtr;
		openvdb::Vec3SGrid::ConstPtr	m_normalGridPtr;
		openvdb::Vec3SGrid::ConstPtr	m_uvGridPtr;

mutable	CountedPool<Impact>	m_vdbImpactPool;
};

} // namespace

#endif /* __vdb_SurfaceVDB_h__ */
